import logging
import re
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Protocol

from iterpy import Arr
from tqdm import tqdm

from .document import Document

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class FileNotRetrievedError(Exception):
    path: Path
    error: Exception

    def __repr__(self) -> str:
        return f"{self.path}: Could not retrieve due to '{self.error}'"


class BaseDocumentSource(Protocol):
    def get_documents(self) -> Sequence[Document]: ...


class MarkdownDocumentSource(BaseDocumentSource):
    """Gets markdown documents. Returns valid markdown."""

    def __init__(self, directory: Path) -> None:
        self.directory = directory

    @staticmethod
    def _replace_wikilinks_with_styling(input_str: str) -> str:
        return input_str.replace("[[", "_").replace("]]", "_")

    @staticmethod
    def _remove_subfolder_from_wiki_links(text: str) -> str:
        ob = r"\[\["
        cb = r"\]\]"
        contents = r"\w|\s|\d|\-|\/"  # Word contents
        matches = re.findall(pattern=rf"{ob}[{contents}]+{cb}", string=text, flags=re.DOTALL)
        for m in matches:
            name = m.replace("[[", "").replace("]]", "").split("/")[-1]
            replacement_str = f"[[{name}]]"
            text = text.replace(m, replacement_str)

        return text

    @staticmethod
    def _only_alias_from_wiki_links(text: str) -> str:
        ob = r"\[\["  # Open bracket
        cb = r"\]\]"  # Close bracket
        word_contents = r"\w|\s|\d|\-"  # Word contents

        regex_pattern = rf"{ob}[{word_contents}|\/\(|\)]+\|[{word_contents}|\/]+{cb}"
        pattern_matches = re.findall(pattern=regex_pattern, string=text, flags=re.DOTALL)

        for match in pattern_matches:
            link_name = (
                re.findall(pattern=r"\|[\w|\s|\d]+\]\]", string=match)[0]
                .replace("|", "")
                .replace("]", "")
                .strip()
            )
            replacement_str = f"[[{link_name}]]"
            text = text.replace(match, replacement_str)

        return text

    @staticmethod
    def _wikilinks_to_markdown(input_str: str) -> str:
        return (
            Arr([input_str])
            .map(MarkdownDocumentSource._only_alias_from_wiki_links)
            .map(MarkdownDocumentSource._remove_subfolder_from_wiki_links)
            .map(MarkdownDocumentSource._replace_wikilinks_with_styling)
        )[0]

    def _get_document_from_file(self, file_path: Path) -> Document | FileNotRetrievedError:
        try:
            try:
                contents = file_path.read_text(encoding="utf8")
            except Exception as e:
                raise Exception(f"Could not read file {file_path}") from e

            try:
                sanitized = self._wikilinks_to_markdown(contents)
            except Exception as e:
                raise Exception(f"Could not sanitize file {file_path}") from e

            return Document(content=sanitized, source_path=file_path)
        except Exception as e:
            log.warning(f"Could not retrieve {file_path}: {e}")
            return FileNotRetrievedError(file_path, e)

    def get_documents(self) -> Sequence[Document]:
        md_files = list(self.directory.rglob("*.md"))

        notes: list[Document | FileNotRetrievedError] = []

        with tqdm(total=len(md_files)) as pbar:
            for filepath in md_files:
                notes.append(self._get_document_from_file(filepath))
                pbar.update(1)

        return [note for note in notes if not isinstance(note, FileNotRetrievedError)]



import os
from pathlib import Path


def get_env(default: str) -> str:
    if os.getenv("GITHUB_ACTIONS"):
        return "GITHUB_ACTIONS"
    return os.getenv("ENV", default)


def get_host_home_dir() -> Path:
    host_home_dir = os.getenv("HOST_HOME")
    home_dir = os.getenv("HOME", ".")
    return Path(host_home_dir) if host_home_dir else Path(home_dir)


def host_input_dir() -> Path:
    host_input_dir = os.getenv("HOST_INPUT_DIR")
    if host_input_dir is None:
        raise Exception("HOST_INPUT_DIR environment variable is not set")
    return Path(host_input_dir)


def in_docker() -> bool:
    return Path("/.dockerenv").exists()



import logging
from dataclasses import dataclass
from pathlib import Path

import pytest

from .document_source import MarkdownDocumentSource


class TestMarkdownIngester:
    def test_happy_path(self, tmp_path: Path):
        with (tmp_path / "test.md").open("w") as f:
            f.write(
                """# Hello World\n
    #anki/tag/test_tag #anki/tag/test_tag2 <!-- #comment_tag -->
    """
            )

        documents = MarkdownDocumentSource(directory=tmp_path).get_documents()

        assert len(documents) == 1
        document = documents[0]
        assert document.title == "test"
        assert document.tags == ["anki/tag/test_tag", "anki/tag/test_tag2", "comment_tag"]

    def test_should_log_error_if_file_not_retrieved(
        self, tmp_path: Path, caplog: pytest.LogCaptureFixture
    ):
        caplog.set_level(logging.WARNING)

        (tmp_path / "succesful.md").write_text("""# I should succeed""")

        failed_path = tmp_path / "failed.md"
        failed_path.write_text("""# I should fail""")
        # Change permission to write only
        # Removing read and execute permissions for owner, group and others
        failed_path.chmod(0o200)

        MarkdownDocumentSource(directory=tmp_path)._get_document_from_file(  # type: ignore[PrivateMethodUsage]
            Path("I do not exist")
        )
        assert "I do not exist" in caplog.records[0].message

    def test_sanitize_to_valid_markdown(self):
        input_str = (
            """Linking to a valid [[Note|Note Alias]], and can handle [[Note2|Multiple Aliases]]"""
        )
        expected_output = """Linking to a valid _Note Alias_, and can handle _Multiple Aliases_"""
        assert (
            MarkdownDocumentSource(directory=Path())._wikilinks_to_markdown(  # type: ignore[PrivateMethodUsage]
                input_str
            )
            == expected_output
        )


@dataclass(frozen=True)
class Ex:
    given: str
    # Required setup

    then: str
    # What the expected result is
    #
    type: str


@pytest.mark.parametrize(
    ("example"),
    [
        Ex("[[N|Alias]]", "_Alias_", "Base"),
        Ex(" [[N|Alias]] ", " _Alias_ ", "Spaces"),
        Ex("[[N/N2|Alias]]", "_Alias_", "Nesting"),
        Ex("[[-|A]]", "_A_", "Dash"),
        Ex("[[swe/Test]]", "_Test_", "Subdir"),
    ],
    ids=lambda x: x.type,
)
def test_sanitise_to_markdown(example: Ex):
    given = MarkdownDocumentSource._wikilinks_to_markdown(example.given)  # type: ignore
    then = example.then
    assert given == then



from memium.destination.ankiconnect.anki_model import AnkiQAModel, Markdown

from ...source.prompt import QAPrompt, QAWithDoc
from ..destination import DestinationPrompt
from .ankiconnect_gateway import NoteInfo


def _edit_button(url: str) -> str:
    return f"""<a href="{url}" style="background-color: #5f0069;
        border: none;
        color: white;
        padding: 0.8em;
        text-align: center;
        text-decoration: none;
        font-size: 0.8em;
        font-family: 'Inter', sans-serif;
        border-radius: 8px;
        opacity: 0.8;
">Obsidian</a>"""


def _extra_field_content(
    source_title: str | None, edit_url: str | None, render_parent_doc: bool
) -> Markdown:
    note_title = (
        f"<span style='opacity: 0.5; font-size: 0.7em; line-height: 0%;'>{source_title}</span>"
        if source_title and render_parent_doc
        else ""
    )
    edit_button_html = _edit_button(edit_url) if edit_url else ""
    return Markdown(
        f"<div class='edit_button' style='text-align: center;'>{edit_button_html}</div>{note_title}"
    )


class AnkiPromptConverter:
    def __init__(self, root_deck: str, deck_prefix: str = "#anki_deck") -> None:
        self.root_deck = root_deck
        self.deck_prefix = deck_prefix

    @staticmethod
    def dummy() -> "AnkiPromptConverter":
        return AnkiPromptConverter(root_deck="FakeBaseDeck")

    def to_destination(self, prompt: QAWithDoc) -> AnkiQAModel:
        deck_in_tags = [tag for tag in prompt.tags if tag.startswith(self.deck_prefix)]
        deck = deck_in_tags[0] if deck_in_tags else self.root_deck

        return AnkiQAModel(
            Question=Markdown(prompt.prompt.question),
            Answer=Markdown(prompt.prompt.answer),
            Extra=_extra_field_content(
                source_title=prompt.parent_doc.title,
                edit_url=prompt.edit_url,
                render_parent_doc=prompt.render_parent_doc,
            ),
            tags=prompt.tags,
            raw_prompt=prompt.prompt,
            root_deck=deck,
        )

    def from_destination(self, note_info: NoteInfo) -> DestinationPrompt:
        if "Question" in note_info.fields and "Answer" in note_info.fields:
            return DestinationPrompt(
                QAPrompt(
                    question=note_info.fields["raw_question"].value,
                    answer=note_info.fields["raw_answer"].value,
                ),
                extra=note_info.fields["Extra"].value,
                destination_id=str(note_info.noteId),
            )

        raise ValueError(f"No Question field in note info: {note_info}")



from inline_snapshot import snapshot

from memium.raw_processors.title_as_answer import TitleAsAnswerProcessor
from memium.source.prompt import QAWithDoc


def test_extract_reversed_definition():
    doc = QAWithDoc.dummy(question="Definition?", answer="A location used for testing.")
    result = TitleAsAnswerProcessor(
        question_matcher="Definition?", reversed_question="Term for '%s'?"
    ).__call__([doc])

    assert len(result) == 2
    assert result[1].prompt.question == snapshot("Term for 'A location used for testing'?")
    assert result[1].prompt.answer == snapshot("DummyPath")


def test_empty_definition_extractor():
    doc = QAWithDoc.dummy(question="No def?", answer="Blah")
    result = TitleAsAnswerProcessor(
        question_matcher="Definition?", reversed_question="Term for '%s'?"
    ).__call__([doc])
    assert len(result) == 1


def test_application_extractor():
    doc = QAWithDoc.dummy(question="Use when?", answer="Preventing coagulation.")
    result = TitleAsAnswerProcessor(
        question_matcher="Use when?", reversed_question="What should we use for '%s'?"
    ).__call__([doc])

    assert len(result) == 2
    assert result[1].prompt.question == snapshot("What should we use for 'Preventing coagulation'?")
    assert result[1].prompt.answer == snapshot("DummyPath")



from inline_snapshot import snapshot

from ...source.prompt import QAWithDoc
from .anki_converter import AnkiPromptConverter


def test_anki_prompt_converter():
    """Tests the AnkiPromptConverter class"""
    input_prompt = QAWithDoc.dummy()
    generated_card = AnkiPromptConverter.dummy().to_destination(input_prompt)

    # Ensure the generated uuid remains stable over time
    assert generated_card.raw_prompt.question == snapshot("FakeQuestion")
    assert generated_card.raw_prompt.answer == snapshot("FakeAnswer")
    assert generated_card.raw_prompt.scheduling_uid == 4875918425
